import torch
import torch.nn as nn
import torch.nn.functional as F
from torch import Tensor
from torch_activation import register_activation


@register_activation
class TanhLinearUnit(nn.Module):
    r"""
    Applies the Tanh Linear Unit activation function:

    :math:`\text{TanhLinearUnit}(z) = \begin{cases} 
    z, & z \geq 0 \\
    \frac{2}{1 + \exp(-z)} - 1, & z < 0 
    \end{cases} = \begin{cases} 
    z, & z \geq 0 \\
    \tanh\left(\frac{z}{2}\right), & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = TanhLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(TanhLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        result[neg_mask] = torch.tanh(x[neg_mask] / 2)
        
        return result


class DualELU(nn.Module):
    r"""
    Applies the Dual ELU activation function:

    :math:`\text{DualELU}(z, z') = \text{ELU}(z) - \text{ELU}(z')`

    Args:
        alpha (float, optional): The alpha value for the ELU formulation. Default: 1.0
        dim (int, optional): The dimension on which to split the input. Default: -1

    Shape:
        - Input: :math:`(*, N, *)` where `*` means any number of dimensions
        - Output: :math:`(*, N/2, *)` where `*` means any number of dimensions

    Examples::

        >>> m = DualELU()
        >>> x = torch.randn(4, 2)
        >>> output = m(x)
    """

    def __init__(self, alpha: float = 1.0, dim: int = -1):
        super(DualELU, self).__init__()
        self.alpha = alpha
        self.dim = dim

    def forward(self, x) -> Tensor:
        dim_size = x.size(self.dim)
        assert dim_size % 2 == 0, f"Dimension {self.dim} must be divisible by 2"
        
        split_size = dim_size // 2
        z, z_prime = torch.split(x, split_size, dim=self.dim)
        
        return F.elu(z, alpha=self.alpha) - F.elu(z_prime, alpha=self.alpha)


@register_activation
class DifferenceELU(nn.Module):
    r"""
    Applies the Difference ELU activation function:

    :math:`\text{DifferenceELU}(z) = \begin{cases} 
    z, & z \geq 0 \\
    a(z\exp(z) - b\exp(bz)), & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Exponential scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = DifferenceELU(a=1.0, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(DifferenceELU, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        neg_x = x[neg_mask]
        result[neg_mask] = self.a * (neg_x * torch.exp(neg_x) - self.b * torch.exp(self.b * neg_x))
        
        return result


@register_activation
class PolynomialLinearUnit(nn.Module):
    r"""
    Applies the Polynomial Linear Unit activation function:

    :math:`\text{PolynomialLinearUnit}(z) = \begin{cases} 
    z, & z \geq 0 \\
    \frac{1}{1 - z} - 1, & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = PolynomialLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(PolynomialLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        neg_x = x[neg_mask]
        
        # Ensure numerical stability by clamping values
        neg_x = torch.clamp(neg_x, min=-0.999)
        result[neg_mask] = 1 / (1 - neg_x) - 1
        
        return result


@register_activation
class InversePolynomialLinearUnit(nn.Module):
    r"""
    Applies the Inverse Polynomial Linear Unit activation function:

    :math:`\text{InversePolynomialLinearUnit}(z) = \begin{cases} 
    z, & z \geq 0 \\
    \frac{1}{1 + |z|^a}, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Power parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = InversePolynomialLinearUnit(a=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(InversePolynomialLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        neg_x = x[neg_mask]
        result[neg_mask] = 1 / (1 + torch.abs(neg_x).pow(self.a))
        
        return result


@register_activation
class PowerLinearUnit(nn.Module):
    r"""
    Applies the Power Linear Unit activation function:

    :math:`\text{PowerLinearUnit}(z) = \begin{cases} 
    z, & z \geq 0 \\
    (1 - z)^{-a} - 1, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Power parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = PowerLinearUnit(a=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(PowerLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        neg_x = x[neg_mask]
        
        # Ensure numerical stability by clamping values
        neg_x = torch.clamp(neg_x, min=-0.999)
        result[neg_mask] = torch.pow(1 - neg_x, -self.a) - 1
        
        return result


@register_activation
class PowerFunctionLinearUnit(nn.Module):
    r"""
    Applies the Power Function Linear Unit activation function:

    :math:`\text{PowerFunctionLinearUnit}(z) = z \cdot \frac{1}{2} \left( 1 + \frac{z}{\sqrt{1 + z^2}} \right)`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = PowerFunctionLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(PowerFunctionLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        return x * 0.5 * (1 + x / torch.sqrt(1 + x.pow(2)))


@register_activation
class FasterPowerFunctionLinearUnit(nn.Module):
    r"""
    Applies the Faster Power Function Linear Unit activation function:

    :math:`\text{FasterPowerFunctionLinearUnit}(z) = \begin{cases} 
    z, & z \geq 0 \\
    z + \frac{z^2}{\sqrt{1 + z^2}}, & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = FasterPowerFunctionLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(FasterPowerFunctionLinearUnit, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = x.clone()
        neg_x = x[neg_mask]
        result[neg_mask] = neg_x + (neg_x.pow(2) / torch.sqrt(1 + neg_x.pow(2)))
        
        return result


@register_activation
class ElasticAdaptivelyParametricCompoundedUnit(nn.Module):
    r"""
    Applies the Elastic Adaptively Parametric Compounded Unit activation function:

    :math:`\text{ElasticAdaptivelyParametricCompoundedUnit}(z_i) = \begin{cases} 
    b_i z_i, & z_i \geq 0 \\
    a_i z_i \cdot \tanh(\ln(1 + \exp(a_{i}z_{i}))), & z_i < 0 
    \end{cases}`

    Args:
        a (float or Tensor, optional): Negative slope parameter. Default: 1.0
        b (float or Tensor, optional): Positive slope parameter. Default: 1.0
        num_parameters (int, optional): Number of parameters if using per-channel parameterization. Default: 1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ElasticAdaptivelyParametricCompoundedUnit(a=0.5, b=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
        
        >>> # Per-channel parameterization
        >>> m = ElasticAdaptivelyParametricCompoundedUnit(num_parameters=3)
        >>> x = torch.randn(3, 5)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, num_parameters: int = 1):
        super(ElasticAdaptivelyParametricCompoundedUnit, self).__init__()
        self.num_parameters = num_parameters
        
        if num_parameters == 1:
            self.a = nn.Parameter(Tensor([a]))
            self.b = nn.Parameter(Tensor([b]))
        else:
            self.a = nn.Parameter(torch.full((num_parameters,), a))
            self.b = nn.Parameter(torch.full((num_parameters,), b))

    def forward(self, x) -> Tensor:
        if self.num_parameters == 1:
            pos_mask = x >= 0
            neg_mask = ~pos_mask
            
            result = torch.zeros_like(x)
            result[pos_mask] = self.b * x[pos_mask]
            
            neg_x = x[neg_mask]
            softplus = torch.log(1 + torch.exp(self.a * neg_x))
            result[neg_mask] = self.a * neg_x * torch.tanh(softplus)
            
            return result
        else:
            # Handle per-channel parameterization
            pos_mask = x >= 0
            neg_mask = ~pos_mask
            
            result = torch.zeros_like(x)
            
            for i in range(self.num_parameters):
                # Apply positive part
                channel_pos_mask = pos_mask.narrow(0, i, 1).squeeze(0)
                if channel_pos_mask.any():
                    result.narrow(0, i, 1)[channel_pos_mask] = self.b[i] * x.narrow(0, i, 1)[channel_pos_mask]
                
                # Apply negative part
                channel_neg_mask = neg_mask.narrow(0, i, 1).squeeze(0)
                if channel_neg_mask.any():
                    neg_x = x.narrow(0, i, 1)[channel_neg_mask]
                    softplus = torch.log(1 + torch.exp(self.a[i] * neg_x))
                    result.narrow(0, i, 1)[channel_neg_mask] = self.a[i] * neg_x * torch.tanh(softplus)
            
            return result


@register_activation
class LipschitzReLU(nn.Module):
    r"""
    Applies the Lipschitz ReLU activation function:

    :math:`\text{LipschitzReLU}(z) = p(z | z > 0) + n(z | z \leq 0)`

    where p and n are positive and negative functions with Lipschitz constant <= 1.

    Args:
        p_fn (callable, optional): Positive function. Default: identity
        n_fn (callable, optional): Negative function. Default: zero

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> # Default implementation (standard ReLU)
        >>> m = LipschitzReLU()
        >>> x = torch.randn(2)
        >>> output = m(x)
        
        >>> # Custom implementation with leaky behavior
        >>> m = LipschitzReLU(p_fn=lambda x: x, n_fn=lambda x: 0.1 * x)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, p_fn=None, n_fn=None):
        super(LipschitzReLU, self).__init__()
        self.p_fn = p_fn if p_fn is not None else lambda x: x
        self.n_fn = n_fn if n_fn is not None else lambda x: torch.zeros_like(x)

    def forward(self, x) -> Tensor:
        pos_mask = x > 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        result[pos_mask] = self.p_fn(x[pos_mask])
        result[neg_mask] = self.n_fn(x[neg_mask])
        
        return result


@register_activation
class ScaledExponentialLinearUnit(nn.Module):
    r"""
    Applies the Scaled Exponential Linear Unit activation function:

    :math:`\text{ScaledExponentialLinearUnit}(z) = \begin{cases} 
    az, & z \geq 0 \\
    ab(\exp(z) - 1), & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.67326
        b (float, optional): Alpha parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ScaledExponentialLinearUnit()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.67326, b: float = 1.0):
        super(ScaledExponentialLinearUnit, self).__init__()
        self.a = a
        self.b = b

    def forward(self, x) -> Tensor:
        return F.selu(x)


@register_activation
class LeakyScaledExponentialLinearUnit(nn.Module):
    r"""
    Applies the Leaky Scaled Exponential Linear Unit activation function:

    :math:`\text{LeakyScaledExponentialLinearUnit}(z) = \begin{cases} 
    az, & z \geq 0 \\
    ab(\exp(z) - 1) + acz, & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Alpha parameter. Default: 1.0
        c (float, optional): Leaky slope. Default: 0.1

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LeakyScaledExponentialLinearUnit(a=1.5, b=1.0, c=0.2)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, c: float = 0.1):
        super(LeakyScaledExponentialLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))
        self.c = nn.Parameter(Tensor([c]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        result[pos_mask] = self.a * x[pos_mask]
        
        neg_x = x[neg_mask]
        result[neg_mask] = self.a * self.b * (torch.exp(neg_x) - 1) + self.a * self.c * neg_x
        
        return result


@register_activation
class ScaledExponentiallyRegularizedLinearUnit(nn.Module):
    r"""
    Applies the Scaled Exponentially Regularized Linear Unit activation function:

    :math:`\text{ScaledExponentiallyRegularizedLinearUnit}(z) = \begin{cases} 
    az, & z \geq 0 \\
    abz\exp(z), & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Regularization parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ScaledExponentiallyRegularizedLinearUnit(a=1.5, b=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(ScaledExponentiallyRegularizedLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        result[pos_mask] = self.a * x[pos_mask]
        
        neg_x = x[neg_mask]
        result[neg_mask] = self.a * self.b * neg_x * torch.exp(neg_x)
        
        return result


@register_activation
class ScaledScaledExponentialLinearUnit(nn.Module):
    r"""
    Applies the Scaled Scaled Exponential Linear Unit activation function:

    :math:`\text{ScaledScaledExponentialLinearUnit}(z) = \begin{cases} 
    az, & z \geq 0 \\
    ab(\exp(cz) - 1), & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0
        b (float, optional): Alpha parameter. Default: 1.0
        c (float, optional): Exponential scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ScaledScaledExponentialLinearUnit(a=1.5, b=1.0, c=0.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, c: float = 1.0):
        super(ScaledScaledExponentialLinearUnit, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))
        self.c = nn.Parameter(Tensor([c]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        result[pos_mask] = self.a * x[pos_mask]
        
        neg_x = x[neg_mask]
        result[neg_mask] = self.a * self.b * (torch.exp(self.c * neg_x) - 1)
        
        return result


@register_activation
class RSigELU(nn.Module):
    r"""
    Applies the RSigELU activation function:

    :math:`\text{RSigELU}(z) = \begin{cases} 
    z \cdot \frac{1}{1 + \exp(-z)} a + z, & 1 < z < \infty \\
    z, & 0 \geq z \geq 1 \\
    a(\exp(z) - 1), & -\infty < z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = RSigELU(a=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(RSigELU, self).__init__()
        self.a = nn.Parameter(Tensor([a]))

    def forward(self, x) -> Tensor:
        result = torch.zeros_like(x)
        
        # Case 1: 1 < z < infinity
        mask1 = x > 1
        result[mask1] = x[mask1] * torch.sigmoid(x[mask1]) * self.a + x[mask1]
        
        # Case 2: 0 <= z <= 1
        mask2 = (x >= 0) & (x <= 1)
        result[mask2] = x[mask2]
        
        # Case 3: -infinity < z < 0
        mask3 = x < 0
        result[mask3] = self.a * (torch.exp(x[mask3]) - 1)
        
        return result


@register_activation
class HardSReLUE(nn.Module):
    r"""
    Applies the Hard SReLUE activation function:

    :math:`\text{HardSReLUE}(z) = \begin{cases} 
    az \cdot \max\left(0, \min\left(1, \frac{z+1}{2} + z\right)\right), & z \geq 0 \\
    a(\exp(z) - 1), & z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HardSReLUE(a=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0):
        super(HardSReLUE, self).__init__()
        self.a = nn.Parameter(Tensor([a]))

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        
        # Positive part
        pos_x = x[pos_mask]
        hard_sigmoid = torch.clamp((pos_x + 1) / 2 + pos_x, 0, 1)
        result[pos_mask] = self.a * pos_x * hard_sigmoid
        
        # Negative part
        neg_x = x[neg_mask]
        result[neg_mask] = self.a * (torch.exp(neg_x) - 1)
        
        return result


@register_activation
class ExponentialLinearSigmoidSquashing(nn.Module):
    r"""
    Applies the Exponential Linear Sigmoid Squashing activation function:

    :math:`\text{ExponentialLinearSigmoidSquashing}(z) = \begin{cases} 
    \frac{z}{1 + \exp(-z)}, & z \geq 0 \\
    \frac{\exp(z) - 1}{1 + \exp(-z)}, & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = ExponentialLinearSigmoidSquashing()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(ExponentialLinearSigmoidSquashing, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        sigmoid = torch.sigmoid(x)
        
        # Positive part
        result[pos_mask] = x[pos_mask] * sigmoid[pos_mask]
        
        # Negative part
        neg_x = x[neg_mask]
        result[neg_mask] = (torch.exp(neg_x) - 1) * sigmoid[neg_mask]
        
        return result


@register_activation
class HardExponentialLinearSigmoidSquashing(nn.Module):
    r"""
    Applies the Hard Exponential Linear Sigmoid Squashing activation function:

    :math:`\text{HardExponentialLinearSigmoidSquashing}(z) = \begin{cases} 
    z \cdot \max\left(0, \min\left(\frac{z+1}{2}, 1\right)\right), & z \geq 0 \\
    (1 + \exp(-z)) \cdot \max\left(0, \min\left(\frac{z+1}{2}, 1\right)\right), & z < 0 
    \end{cases}`

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HardExponentialLinearSigmoidSquashing()
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self):
        super(HardExponentialLinearSigmoidSquashing, self).__init__()

    def forward(self, x) -> Tensor:
        pos_mask = x >= 0
        neg_mask = ~pos_mask
        
        result = torch.zeros_like(x)
        hard_sigmoid = torch.clamp((x + 1) / 2, 0, 1)
        
        # Positive part
        result[pos_mask] = x[pos_mask] * hard_sigmoid[pos_mask]
        
        # Negative part
        neg_x = x[neg_mask]
        result[neg_mask] = (1 + torch.exp(-neg_x)) * hard_sigmoid[neg_mask]
        
        return result


@register_activation
class RSigELUD(nn.Module):
    r"""
    Applies the RSigELUD activation function:

    :math:`\text{RSigELUD}(z) = \begin{cases} 
    z \cdot \frac{1}{1 + \exp(-z)} a + z, & 1 < z < \infty \\
    z, & 0 \leq z \leq 1 \\
    b(\exp(z) - 1), & -\infty < z < 0 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter for z > 1. Default: 1.0
        b (float, optional): Scale parameter for z < 0. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = RSigELUD(a=1.5, b=0.8)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(RSigELUD, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        result = torch.zeros_like(x)
        
        # Case 1: 1 < z < infinity
        mask1 = x > 1
        result[mask1] = x[mask1] * torch.sigmoid(x[mask1]) * self.a + x[mask1]
        
        # Case 2: 0 <= z <= 1
        mask2 = (x >= 0) & (x <= 1)
        result[mask2] = x[mask2]
        
        # Case 3: -infinity < z < 0
        mask3 = x < 0
        result[mask3] = self.b * (torch.exp(x[mask3]) - 1)
        
        return result


@register_activation 
class LSReLU(nn.Module):
    r"""
    Applies the LSReLU activation function:

    :math:`\text{LSReLU}(z) = \begin{cases} 
    \frac{z}{1 + |z|}, & z \leq 0 \\
    z, & 0 \leq z \leq b \\
    \log(az + 1) + |\log(ab + 1) - b|, & z \geq b 
    \end{cases}`

    Args:
        a (float, optional): Scale parameter for z > b. Default: 1.0
        b (float, optional): Threshold parameter. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LSReLU(a=0.5, b=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0):
        super(LSReLU, self).__init__()
        self.a = nn.Parameter(Tensor([a]))
        self.b = nn.Parameter(Tensor([b]))

    def forward(self, x) -> Tensor:
        result = torch.zeros_like(x)
        
        # Case 1: z <= 0
        mask1 = x <= 0
        neg_x = x[mask1]
        result[mask1] = neg_x / (1 + torch.abs(neg_x))
        
        # Case 2: 0 <= z <= b
        mask2 = (x > 0) & (x <= self.b)
        result[mask2] = x[mask2]
        
        # Case 3: z >= b
        mask3 = x > self.b
        log_term = torch.log(self.a * x[mask3] + 1)
        offset = torch.abs(torch.log(self.a * self.b + 1) - self.b)
        result[mask3] = log_term + offset
        
        return result


import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor

from torch_activation import register_activation

@register_activation
class CosLU(nn.Module):
    r"""
    Applies the Cosine Linear Unit function:

    :math:`\text{CosLU}(x) = (x + a \cdot \cos(b \cdot x)) \cdot \sigma(x)`

     See: https://doi.org/10.20944/preprints202301.0463.v1

    Args:
        a (float, optional): Scaling factor for the cosine term. Default is 1.0.
        b (float, optional): Frequency factor for the cosine term. Default is 1.0.
        inplace (bool, optional): can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/CosLU.png

    Examples::

        >>> m = CosLU(alpha=2.0, beta=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = CosLU(inplace=True)
        >>> x = torch.randn(2, 3, 4)
        >>> m(x)
    """

    def __init__(self, a: float = 1.0, b: float = 1.0, inplace: bool = False):
        super(CosLU, self).__init__()
        self.alpha = nn.Parameter(Tensor([a]))
        self.beta = nn.Parameter(Tensor([b]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        return self._forward_inplace(x) if self.inplace else self._forward(x)

    def _forward(self, x):
        result = x + self.alpha * torch.cos(self.beta * x)
        result *= torch.sigmoid(x)
        return result

    def _forward_inplace(self, x):
        s_x = torch.sigmoid(x)
        x.add_(self.alpha * torch.cos(self.beta * x))
        x.mul_(s_x)
        return x

@register_activation
class LAAF(nn.Module):
    r"""
    Applies the Locally Adaptive Activation Function (LAAF):

    :math:`\text{LAAF}(x) = g(a \cdot x)`

    where :math:`a` is a trainable parameter for each neuron and :math:`g` is any activation function.

    See: https://doi.org/10.1016/j.cma.2020.113028

    Args:
        activation (str, optional): The activation function to use. Options: 'sigmoid', 'tanh', 'relu', 'leaky_relu'. Default: 'sigmoid'
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0
        leaky_slope (float, optional): Leakiness parameter for LeakyReLU. Default: 0.01
        fixed_n (float, optional): Fixed parameter to accelerate convergence. If > 1, applies g(n*a*x). Default: 1.0
        inplace (bool, optional): Can optionally do the operation in-place when possible. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LAAF(activation='tanh', a_init=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = LAAF(activation='relu', fixed_n=2.0, inplace=True)
        >>> x = torch.randn(2, 3, 4)
        >>> m(x)
    """

    def __init__(self, activation: str = 'sigmoid', a_init: float = 1.0, 
                 leaky_slope: float = 0.01, fixed_n: float = 1.0, inplace: bool = False):
        super(LAAF, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.activation = activation.lower()
        self.leaky_slope = leaky_slope
        self.fixed_n = fixed_n
        self.inplace = inplace
        
        if self.activation not in ['sigmoid', 'tanh', 'relu', 'leaky_relu']:
            raise ValueError(f"Unsupported activation: {activation}. Choose from 'sigmoid', 'tanh', 'relu', 'leaky_relu'")

    def forward(self, x) -> Tensor:
        scaled_x = self.fixed_n * self.a * x
        
        if self.activation == 'sigmoid':
            return torch.sigmoid(scaled_x)
        elif self.activation == 'tanh':
            return torch.tanh(scaled_x)
        elif self.activation == 'relu':
            return F.relu(scaled_x, inplace=self.inplace)
        elif self.activation == 'leaky_relu':
            return F.leaky_relu(scaled_x, negative_slope=self.leaky_slope, inplace=self.inplace)


@register_activation
class AdaptiveSlopeTanh(nn.Module):
    r"""
    Applies the Adaptive Slope Hyperbolic Tangent function:

    :math:`\text{AdaptiveSlopeTanh}(x) = \tanh(a \cdot x)`

    where :math:`a` is a trainable parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = AdaptiveSlopeTanh(a_init=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0):
        super(AdaptiveSlopeTanh, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))

    def forward(self, x) -> Tensor:
        return torch.tanh(self.a * x)


@register_activation
class PSTanh(nn.Module):
    r"""
    Applies the Parametric Scaled Hyperbolic Tangent function:

    :math:`\text{PSTanh}(x) = x \cdot a \cdot (1 + \tanh(b \cdot x))`

    where :math:`a` and :math:`b` are trainable parameters.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = PSTanh(a_init=2.0, b_init=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0, b_init: float = 1.0):
        super(PSTanh, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))

    def forward(self, x) -> Tensor:
        return x * self.a * (1 + torch.tanh(self.b * x))


@register_activation
class SSinH(nn.Module):
    r"""
    Applies the Scaled Sine-Hyperbolic function:

    :math:`\text{SSinH}(x) = a \cdot \sinh(b \cdot x)`

    where :math:`a` and :math:`b` are trainable parameters.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SSinH(a_init=2.0, b_init=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0, b_init: float = 1.0):
        super(SSinH, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))

    def forward(self, x) -> Tensor:
        return self.a * torch.sinh(self.b * x)


@register_activation
class SExp(nn.Module):
    r"""
    Applies the Scaled Exponential function:

    :math:`\text{SExp}(x) = a \cdot (\exp(b \cdot x) - 1)`

    where :math:`a` and :math:`b` are trainable parameters.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = SExp(a_init=2.0, b_init=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0, b_init: float = 1.0):
        super(SExp, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))

    def forward(self, x) -> Tensor:
        return self.a * (torch.exp(self.b * x) - 1)


@register_activation
class LAU(nn.Module):
    r"""
    Applies the Logmoid Activation Unit function:

    :math:`\text{LAU}(x) = x \cdot \ln(1 + a \cdot \sigma(b \cdot x))`

    where :math:`a` and :math:`b` are trainable parameters and :math:`\sigma` is the sigmoid function.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = LAU(a_init=2.0, b_init=1.5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0, b_init: float = 1.0):
        super(LAU, self).__init__()
        self.a = nn.Parameter(Tensor([a_init]))
        self.b = nn.Parameter(Tensor([b_init]))

    def forward(self, x) -> Tensor:
        return x * torch.log(1 + self.a * torch.sigmoid(self.b * x))


@register_activation
class AGumb(nn.Module):
    r"""
    Applies the Adaptive Gumbel function:

    :math:`\text{AGumb}(x) = 1 - (1 + a \cdot \exp(x))^{-1}`

    where :math:`a` is a trainable positive parameter.

    Args:
        a_init (float, optional): Initial value for the trainable parameter a. Default: 1.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = AGumb(a_init=2.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a_init: float = 1.0):
        super(AGumb, self).__init__()
        # Using softplus to ensure a is positive
        self.a_raw = nn.Parameter(Tensor([a_init]))

    def forward(self, x) -> Tensor:
        # Ensure a is positive using softplus
        a = F.softplus(self.a_raw)
        return 1 - (1 + a * torch.exp(x))**(-1)



import os
import importlib
import inspect
import sys

# Get the directory of the current file
current_dir = os.path.dirname(os.path.abspath(__file__))

# Get all Python files in the directory (excluding __init__.py)
python_files = [f[:-3] for f in os.listdir(current_dir) 
                if f.endswith('.py') and f != '__init__.py']

# Dictionary to store all classes
__all__ = []

# Import all classes from each file
for module_name in python_files:
    # Import the module
    module = importlib.import_module(f"torch_activation.adaptive.{module_name}")
    
    # Get all classes from the module
    for name, obj in inspect.getmembers(module, inspect.isclass):
        # Check if the class is defined in this module (not imported)
        if obj.__module__ == f"torch_activation.adaptive.{module_name}":
            # Add the class to the current namespace
            globals()[name] = obj
            __all__.append(name)

# Clean up namespace
del os, importlib, inspect, sys, current_dir, python_files, module_name, name, obj, module



import torch
import torch.nn as nn
import torch.nn.functional as F

from torch import Tensor
from typing import Iterable


class LinComb(nn.Module):
    r"""
    Applies the LinComb activation function:

    :math:`\text{LinComb}(x) = \sum_{i=1}^{n} w_i \cdot F_i(x)`

     See: https://doi.org/10.20944/preprints202301.0463.v1

    Args:
        activations (Iterable[nn.Module]): List of activation functions. Default: [nn.ReLU, nn.Sigmoid, nn.Tanh, nn.Softsign]

    Shape:
        - Input: :math:`(*)` where :math:`*` means any number of additional dimensions.
        - Output: :math:`(*)`

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/LinComb.png

    Examples::

        >>> activations = [nn.ReLU(), nn.Sigmoid()]
        >>> m = LinComb(activation_functions)
        >>> input = torch.randn(10)
        >>> output = m(input)
    """

    def __init__(
        self,
        activations: Iterable[nn.Module] = [
            nn.ReLU(),
            nn.Sigmoid(),
            nn.Tanh(),
            nn.Softsign(),
        ],
    ):
        super(LinComb, self).__init__()
        self.activations = nn.ModuleList(activations)
        self.weights = nn.Parameter(Tensor(len(activations)))

        self.weights.data.uniform_(-1, 1)

    def forward(self, input) -> Tensor:
        activations = [
            self.weights[i] * self.activations[i](input)
            for i in range(len(self.activations))
        ]
        return torch.sum(torch.stack(activations), dim=0)


class NormLinComb(nn.Module):
    r"""
    Applies the LinComb activation function:

    :math:`\text{NormLinComb}(x) = \frac{\sum_{i=1}^{n} w_i \cdot F_i(x)}{\|\|W\|\|}`

     See: https://doi.org/10.20944/preprints202301.0463.v1

    Args:
        activations (Iterable[nn.Module]): List of activation functions. Default: [nn.ReLU, nn.Sigmoid, nn.Tanh, nn.Softsign]

    Shape:
        - Input: :math:`(*)` where :math:`*` means any number of additional dimensions.
        - Output: :math:`(*)`

    Here is a plot of the function and its derivative:

    .. image:: ../images/activation_images/NormLinComb.png

    Examples::

        >>> activations = [nn.ReLU, nn.Sigmoid]
        >>> m = NormLinComb(activation_functions)
        >>> input = torch.randn(10)
        >>> output = m(input)
    """

    def __init__(
        self,
        activations: Iterable[nn.Module] = [
            nn.ReLU(),
            nn.Sigmoid(),
            nn.Tanh(),
            nn.Softsign(),
        ],
    ):
        super(NormLinComb, self).__init__()
        self.activations = nn.ModuleList(activations)
        self.weights = nn.Parameter(torch.Tensor(len(activations)))

        self.weights.data.uniform_(-1, 1)

    def forward(self, input) -> torch.Tensor:
        activations = [
            self.weights[i] * self.activations[i](input)
            for i in range(len(self.activations))
        ]
        output = torch.sum(torch.stack(activations), dim=0)
        return output / torch.norm(output)




import torch
import torch.nn as nn
from torch import Tensor
from torch_activation import register_activation

class HCAF(nn.Module):
    r"""
    Applies the Hybrid Chaotic Activation Function:

    .. math::
        a_i = \sigma(z_i)
        
        c_{i,1} = ra_i(1 - a_i)
        
        c_{i,j} = rc_{i,j-1}(1 - c_{i,j-1})

    where :math:`\sigma(z_i)` is the logistic sigmoid and :math:`r = 4` by default.

    Args:
        r (float, optional): Chaotic parameter. Default: ``4.0``
        iterations (int, optional): Number of iterations for the chaotic map. Default: ``3``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.HCAF()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.HCAF(r=3.9, iterations=5)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, r: float = 4.0, iterations: int = 3):
        super().__init__()
        self.r = r
        self.iterations = iterations

    def forward(self, x: Tensor) -> Tensor:
        # Initial sigmoid activation
        a = torch.sigmoid(x)
        
        # First chaotic iteration
        c = self.r * a * (1 - a)
        
        # Additional chaotic iterations
        for _ in range(1, self.iterations):
            c = self.r * c * (1 - c)
            
        return c


@register_activation
class FCAF_Hidden(nn.Module):
    r"""
    Applies the Fusion of Chaotic Activation Function for hidden units:

    .. math::
        f(z_{i+1}) = rz_i(1 - z_i) + z_i + a - \frac{b}{2\pi} \sin(2\pi z_i)

    Args:
        r (float, optional): Chaotic parameter. Default: ``4.0``
        a (float, optional): Linear shift parameter. Default: ``0.0``
        b (float, optional): Sinusoidal amplitude parameter. Default: ``0.5``
        iterations (int, optional): Number of iterations for the chaotic map. Default: ``1``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.FCAF_Hidden()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.FCAF_Hidden(r=3.9, a=0.1, b=0.3, iterations=2)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, r: float = 4.0, a: float = 0.0, b: float = 0.5, iterations: int = 1):
        super().__init__()
        self.r = r
        self.a = a
        self.b = b
        self.iterations = iterations

    def forward(self, x: Tensor) -> Tensor:
        # Normalize input to [0,1] range for chaotic map stability
        z = torch.sigmoid(x)
        
        for _ in range(self.iterations):
            # Apply the chaotic map
            z = self.r * z * (1 - z) + z + self.a - (self.b / (2 * torch.pi)) * torch.sin(2 * torch.pi * z)
            
        return z


@register_activation
class FCAF_Output(nn.Module):
    r"""
    Applies the Fusion of Chaotic Activation Function for output units:

    .. math::
        f(z_{i+1}) = rz_i(1 - z_i) + z_i + a - \frac{b}{2\pi} \sin(2\pi z_i) + \exp(-cz_i^2) + d

    Args:
        r (float, optional): Chaotic parameter. Default: ``4.0``
        a (float, optional): Linear shift parameter. Default: ``0.0``
        b (float, optional): Sinusoidal amplitude parameter. Default: ``0.5``
        c (float, optional): Gaussian width parameter. Default: ``1.0``
        d (float, optional): Constant shift parameter. Default: ``0.0``
        iterations (int, optional): Number of iterations for the chaotic map. Default: ``1``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.FCAF_Output()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.FCAF_Output(r=3.9, a=0.1, b=0.3, c=2.0, d=0.1, iterations=2)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, r: float = 4.0, a: float = 0.0, b: float = 0.5, 
                 c: float = 1.0, d: float = 0.0, iterations: int = 1):
        super().__init__()
        self.r = r
        self.a = a
        self.b = b
        self.c = c
        self.d = d
        self.iterations = iterations

    def forward(self, x: Tensor) -> Tensor:
        # Normalize input to [0,1] range for chaotic map stability
        z = torch.sigmoid(x)
        
        for _ in range(self.iterations):
            # Apply the chaotic map with additional terms
            z = (self.r * z * (1 - z) + 
                 z + self.a - 
                 (self.b / (2 * torch.pi)) * torch.sin(2 * torch.pi * z) + 
                 torch.exp(-self.c * z * z) + 
                 self.d)
            
        return z


@register_activation
class CCAF(nn.Module):
    r"""
    Applies the Cascade Chaotic Activation Function:

    .. math::
        f(z_{i+1}) = a \cdot \sin(\pi \cdot b \cdot \sin(\pi z_i))

    where :math:`a, b \in [0, 1]`.

    Args:
        a (float, optional): Amplitude parameter. Default: ``0.5``
        b (float, optional): Inner sine scaling parameter. Default: ``0.5``
        iterations (int, optional): Number of iterations for the chaotic map. Default: ``1``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = torch_activation.CCAF()
        >>> x = torch.randn(2)
        >>> output = m(x)

        >>> m = torch_activation.CCAF(a=0.8, b=0.7, iterations=3)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, a: float = 0.5, b: float = 0.5, iterations: int = 1):
        super().__init__()
        assert 0 <= a <= 1, "Parameter 'a' must be in the range [0, 1]"
        assert 0 <= b <= 1, "Parameter 'b' must be in the range [0, 1]"
        self.a = a
        self.b = b
        self.iterations = iterations

    def forward(self, x: Tensor) -> Tensor:
        # Normalize input to [-1,1] range for sine stability
        z = torch.tanh(x)
        
        for _ in range(self.iterations):
            # Apply the cascade chaotic map
            z = self.a * torch.sin(torch.pi * self.b * torch.sin(torch.pi * z))
            
        return z


import torch
import torch.nn as nn
import torch.nn.functional as F
import math
from torch import Tensor

from torch_activation import register_activation

@register_activation
class MeLU(nn.Module):
    r"""
    Applies the Mexican ReLU (MeLU) function:

    :math:`\text{MeLU}(z_i) = \text{PReLU}(z_i) + \sum_{j=1}^{k-1} a_{i,j} \phi_{b_j c_j}(z_i)`

    where:
    :math:`\phi_{b_j c_j}(z_i) = \max(c_j - |z_i - b_j|, 0)`

    and :math:`a_{i,j}` are trainable parameters, :math:`b_j` and :math:`c_j` are fixed constants.

    Args:
        k (int, optional): Number of trainable parameters (k-1 for the sum and one for PReLU). Default: 4
        init_negative_slope (float, optional): Initial value for the PReLU negative slope. Default: 0.01
        init_a (float, optional): Initial value for the trainable parameters a. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MeLU(k=4)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, k: int = 4, init_negative_slope: float = 0.01, init_a: float = 0.0):
        super(MeLU, self).__init__()
        self.k = k
        
        # PReLU parameter
        self.prelu_weight = nn.Parameter(Tensor([init_negative_slope]))
        
        # Trainable parameters a_i,j
        self.a = nn.Parameter(torch.full((k-1,), init_a))
        
        # Fixed parameters b_j and c_j
        # Initialize them recursively as mentioned in the paper
        self.b = torch.zeros(k-1)
        self.c = torch.zeros(k-1)
        
        # Initialize b_j and c_j recursively
        # This is a simple initialization scheme; the paper may have a more specific one
        for j in range(k-1):
            self.b[j] = j * 2.0 / (k-1) - 1.0  # Spread between -1 and 1
            self.c[j] = 1.0 / (j+1)            # Decreasing values

    def forward(self, x) -> Tensor:
        # PReLU part
        prelu_out = F.prelu(x, self.prelu_weight)
        
        # Sum part
        sum_part = torch.zeros_like(x)
        for j in range(self.k-1):
            # Calculate phi_b_j,c_j(z_i)
            phi = torch.clamp(self.c[j] - torch.abs(x - self.b[j]), min=0.0)
            sum_part += self.a[j] * phi
        
        return prelu_out + sum_part


@register_activation
class MMeLU(nn.Module):
    r"""
    Applies the Modified Mexican ReLU (MMeLU) function:

    :math:`\text{MMeLU}(z_i) = a_i \cdot \max(b_i - |z_i - c_i|, 0) + (1 - a_i) \cdot \text{ReLU}(z_i)`

    where :math:`a_i \in [0, 1]`, :math:`b_i \in \mathbb{R}^+`, and :math:`c_i \in \mathbb{R}` are trainable parameters.

    Args:
        init_a (float, optional): Initial value for parameter a. Default: 0.5
        init_b (float, optional): Initial value for parameter b. Default: 1.0
        init_c (float, optional): Initial value for parameter c. Default: 0.0
        inplace (bool, optional): Can optionally do the operation in-place for ReLU. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = MMeLU(init_a=0.5, init_b=1.0, init_c=0.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, init_a: float = 0.5, init_b: float = 1.0, init_c: float = 0.0, inplace: bool = False):
        super(MMeLU, self).__init__()
        self.a_raw = nn.Parameter(Tensor([init_a]))
        self.b_raw = nn.Parameter(Tensor([init_b]))
        self.c = nn.Parameter(Tensor([init_c]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        # Constrain a to [0, 1] using sigmoid
        a = torch.sigmoid(self.a_raw)
        
        # Constrain b to be positive using softplus
        b = F.softplus(self.b_raw)
        
        # First term: a * max(b - |z - c|, 0)
        first_term = a * torch.clamp(b - torch.abs(x - self.c), min=0.0)
        
        # Second term: (1 - a) * ReLU(z)
        second_term = (1 - a) * F.relu(x, inplace=self.inplace)
        
        return first_term + second_term


@register_activation
class GaLU(nn.Module):
    r"""
    Applies the Gaussian ReLU (GaLU) function:

    :math:`\text{GaLU}(z_i) = \text{PReLU}(z_i) + \sum_{j=1}^{k-1} a_{i,j} \phi_{b_j c_j}(z_i)`

    where:
    :math:`\phi_{b_j c_j}(z_i) = \max(c_j - |z_i - b_j|, 0) + \min(|z - b_j - 2c_j| - c_j, 0)`

    and :math:`a_{i,j}` are trainable parameters, :math:`b_j` and :math:`c_j` are fixed constants.

    Args:
        k (int, optional): Number of trainable parameters (k-1 for the sum and one for PReLU). Default: 4
        init_negative_slope (float, optional): Initial value for the PReLU negative slope. Default: 0.01
        init_a (float, optional): Initial value for the trainable parameters a. Default: 0.0

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = GaLU(k=4)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, k: int = 4, init_negative_slope: float = 0.01, init_a: float = 0.0):
        super(GaLU, self).__init__()
        self.k = k
        
        # PReLU parameter
        self.prelu_weight = nn.Parameter(Tensor([init_negative_slope]))
        
        # Trainable parameters a_i,j
        self.a = nn.Parameter(torch.full((k-1,), init_a))
        
        # Fixed parameters b_j and c_j
        # Initialize them similarly to MeLU
        self.b = torch.zeros(k-1)
        self.c = torch.zeros(k-1)
        
        # Initialize b_j and c_j
        for j in range(k-1):
            self.b[j] = j * 2.0 / (k-1) - 1.0  # Spread between -1 and 1
            self.c[j] = 1.0 / (j+1)            # Decreasing values

    def forward(self, x) -> Tensor:
        # PReLU part
        prelu_out = F.prelu(x, self.prelu_weight)
        
        # Sum part
        sum_part = torch.zeros_like(x)
        for j in range(self.k-1):
            # Calculate phi_b_j,c_j(z_i) for GaLU
            term1 = torch.clamp(self.c[j] - torch.abs(x - self.b[j]), min=0.0)
            term2 = torch.clamp(torch.abs(x - self.b[j] - 2*self.c[j]) - self.c[j], max=0.0)
            phi = term1 + term2
            sum_part += self.a[j] * phi
        
        return prelu_out + sum_part


@register_activation
class HardSwish(nn.Module):
    r"""
    Applies the Hard-Swish function:

    :math:`\text{HardSwish}(z_i) = 2z_i \cdot \max(0, \min(0.2b_i z_i + 0.5, 1))`

    where :math:`b_i` is a trainable parameter.

    Args:
        b_init (float, optional): Initial value for the trainable parameter b. Default: 1.0
        inplace (bool, optional): Can optionally do the operation in-place. Default: ``False``

    Shape:
        - Input: :math:`(*)`, where :math:`*` means any number of dimensions.
        - Output: :math:`(*)`, same shape as the input.

    Examples::

        >>> m = HardSwish(b_init=1.0)
        >>> x = torch.randn(2)
        >>> output = m(x)
    """

    def __init__(self, b_init: float = 1.0, inplace: bool = False):
        super(HardSwish, self).__init__()
        self.b = nn.Parameter(Tensor([b_init]))
        self.inplace = inplace

    def forward(self, x) -> Tensor:
        # Calculate hard sigmoid part: max(0, min(0.2*b*x + 0.5, 1))
        hard_sigmoid = torch.clamp(0.2 * self.b * x + 0.5, min=0.0, max=1.0)
        
        # Multiply by 2*x
        if self.inplace and x.is_floating_point():
            result = x.mul_(2).mul_(hard_sigmoid)
            return result
        else:
            return 2 * x * hard_sigmoid
